package com.pointrlabs.cordova.pointr;

import android.Manifest;
import android.content.Intent;

import com.pointrlabs.cordova.pointr.core.activity.MapActivity;

import org.apache.cordova.CallbackContext;
import org.apache.cordova.CordovaPlugin;
import org.json.JSONArray;
import org.json.JSONException;

import com.pointrlabs.core.geometry.GeoPoint;
import com.pointrlabs.core.geometry.Geometry;
import com.pointrlabs.core.geometry.GeometryType;
import com.pointrlabs.core.management.PermissionManager;
import com.pointrlabs.core.management.PoiManager;
import com.pointrlabs.core.management.models.Building;
import com.pointrlabs.core.management.models.Level;
import com.pointrlabs.core.management.models.PTRParams;
import com.pointrlabs.core.management.models.PointrEnvironment;
import com.pointrlabs.core.nativecore.wrappers.Plog;
import com.google.gson.Gson;
import com.pointrlabs.core.management.Pointr;
import com.pointrlabs.core.management.PositionManager;
import com.pointrlabs.core.pathfinding.PathFindingMode;
import com.pointrlabs.core.pathfinding.PathManager;
import com.pointrlabs.core.poi.models.Poi;
import com.pointrlabs.core.poi.models.PoiContainer;
import com.pointrlabs.core.positioning.model.CalculatedLocation;

import com.pointrlabs.cordova.pointr.core.models.Location;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import com.pointrlabs.core.management.DataManager;
import com.pointrlabs.core.management.interfaces.PointrListener;
import com.pointrlabs.core.management.models.DataType;
import com.pointrlabs.core.management.models.ErrorMessage;
import com.pointrlabs.core.management.models.Site;
import com.pointrlabs.core.positioning.model.PositioningTypes;
import com.pointrlabs.core.site.SiteManager;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;


public class PointrPlugin extends CordovaPlugin {
  ExecutorService backgroundThread = Executors.newCachedThreadPool();
  String clientId;
  String licenseKey;
  String baseUrl;
  String environment;
  Integer logLevel = 2;

  PermissionManager permissionManager;
  public static boolean shouldRequestPermissionsAtStartup = true;

  @Override
  public boolean execute(String action, JSONArray args, CallbackContext callbackContext) {
    switch (action) {
      case "shouldRequestPermissionsAtStartup": {
        try {
          shouldRequestPermissionsAtStartup = args.getBoolean(0);
        } catch (JSONException ignored) { }
        callbackContext.success();
        return true;
      }
      case "initialize": {
        try {
          licenseKey = args.getString(0);
          environment = args.getString(1);
        } catch (JSONException ex) {
          Plog.e("PointrPlugin ->  Critical error! Parameters for initializing Pointr not given properly.");
          callbackContext.error("PointrPlugin ->  Critical error! Could not initialize Pointr Exception: " + ex.getMessage());
          return true;
        }
        try {
          logLevel = args.getInt(2);
        } catch (JSONException ex) {
          Plog.w("PointrPlugin -> Log level not given for initializing Pointr. It will default to INFO");
        }
        callbackContext.success();
        return true;
      }
      case "initializeWithoutContent": {
        try {
          clientId = args.getString(0);
          licenseKey = args.getString(1);
          baseUrl = args.getString(2);
        } catch (JSONException ex) {
          Plog.e("PointrPlugin ->  Critical error! Parameters for initializing Pointr not given properly.");
          callbackContext.error("PointrPlugin ->  Critical error! Could not initialize Pointr Exception: " + ex.getMessage());
          return true;
        }
        try {
          logLevel = args.getInt(3);
        } catch (JSONException ex) {
          Plog.w("PointrPlugin -> Log level not given for initializing Pointr. It will default to INFO");
        }
        callbackContext.success();
        return true;
      }
      case "start": {
        backgroundThread.execute(() -> this.startPointr(callbackContext));
        return true;
      }
      case "stop": {
        backgroundThread.execute(() -> {
          // This is not included in the V1.0
          //this.stopPointr();
        });
        return true;
      }
      case "showPoiDetails": {
        final String siteExternalIdentifier;
        final String poiInternalIdentifier;
        try {
          siteExternalIdentifier = args.getString(0);
          poiInternalIdentifier = args.getString(1);
        } catch (JSONException ex) {
          Plog.e("PointrPlugin ->  Could not retrieve the arguments for showing Poi details");
          callbackContext.error("PointrPlugin ->  Could not retrieve the arguments for showing Poi details");
          return true;
        }
        backgroundThread.execute(() -> this.showPoiDetails(siteExternalIdentifier, poiInternalIdentifier, callbackContext));
        return true;
      }
      case "showPathfindingToPoi": {
        final String siteExternalIdentifier;
        final String poiInternalIdentifier;
        try {
          siteExternalIdentifier = args.getString(0);
          poiInternalIdentifier = args.getString(1);
        } catch (JSONException ex) {
          Plog.e("PointrPlugin ->  Could not retrieve the arguments for starting path to Poi");
          callbackContext.error("PointrPlugin ->  Could not retrieve the arguments for pathfinding poi details");
          return true;
        }
        backgroundThread.execute(() -> this.showPathfindingToPoi(siteExternalIdentifier, poiInternalIdentifier, callbackContext));
        return true;
      }
      case "showPathfindingBetweenPois": {
        final String siteExternalIdentifier;
        final String sourcePoiInternalIdentifier;
        final String destinationPoiInternalIdentifier;
        try {
          siteExternalIdentifier = args.getString(0);
          sourcePoiInternalIdentifier = args.getString(1);
          destinationPoiInternalIdentifier = args.getString(2);
        } catch (JSONException ex) {
          Plog.e("PointrPlugin ->  Could not retrieve siteExternalIdentifier, sourcePoiInternalIdentifier, destinationPoiInternalIdentifier information.");
          callbackContext.error(ex.getMessage());
          return true;
        }
        backgroundThread.execute(() -> this.showPathBetweenPois(siteExternalIdentifier, sourcePoiInternalIdentifier, destinationPoiInternalIdentifier, callbackContext));
        return true;
      }
      case "clearPath": {
        this.clearPath();
        return true;
      }
      case "showBuilding": {
        final String siteExternalIdentifier;
        final String buildingExternalIdentifier;
        try {
          siteExternalIdentifier = args.getString(0);
          buildingExternalIdentifier = args.getString(1);
        } catch (JSONException ex) {
          Plog.e("PointrPlugin ->  Could not retrieve the arguments for showing building");
          callbackContext.error(ex.getMessage());
          return true;
        }
        backgroundThread.execute(() -> this.showMapForBuilding(siteExternalIdentifier, buildingExternalIdentifier, callbackContext));
        callbackContext.success();
        return true;
      }
      case "showSite": {
        final String siteExternalIdentifier;
        try {
          siteExternalIdentifier = args.getString(0);
        } catch (JSONException ex) {
          Plog.e("PointrPlugin ->  Could not retrieve site external identifier. Cannot show site");
          callbackContext.error(ex.getMessage());
          return true;
        }
        backgroundThread.execute(() -> this.showMapForSite(siteExternalIdentifier, callbackContext));
        callbackContext.success();
        return true;
      }
      case "showLevel": {
        final String siteExternalIdentifier;
        final String buildingExternalIdentifier;
        final int levelIndex;
        try {
          siteExternalIdentifier = args.getString(0);
          buildingExternalIdentifier = args.getString(1);
          levelIndex = args.getInt(2);
        } catch (JSONException ex) {
          Plog.e("PointrPlugin ->  Could not retrieve building and level index. Cannot show level");
          callbackContext.error(ex.getMessage());
          return true;
        }
        backgroundThread.execute(() -> this.showMapForLevel(siteExternalIdentifier, buildingExternalIdentifier, levelIndex, callbackContext));
        callbackContext.success();
        return true;
      }
      case "showMap": {
        backgroundThread.execute(this::showMap);
        callbackContext.success();
        return true;
      }
      case "highlightPoi": {
        final String siteExternalIdentifier;
        try {
          siteExternalIdentifier = args.getString(0);
        } catch (JSONException e) {
          callbackContext.error(e.getMessage());
          return true;
        }
        final String poiId;
        try {
          poiId = args.getString(1);
        } catch (JSONException e) {
          callbackContext.error(e.getMessage());
          return true;
        }
        if (poiId == null || siteExternalIdentifier == null) {
          Plog.w("PointrPlugin ->  Could not get parameters needed for zooming in to Poi." +
            "Please provide a valid poi id and site external identifier.");
          callbackContext.error("PointrPlugin ->  Wrong parameters for building id and poi id");
          return true;
        }
        Plog.v("PointrPlugin ->  Poi id retrieved: " + poiId);
        Plog.v("PointrPlugin ->  Site id retrieved: " + siteExternalIdentifier);
        backgroundThread.execute(() -> this.highlightPoi(siteExternalIdentifier, poiId, callbackContext));
        return true;
      }
      case "getCurrentLocation": {
        final String typeOfClosestPoi;
        String tempClosestPoi;
        final Integer finalOptionalTimeout;
        Integer optionalTimeout;
        final Double finalOptionalThreshold;
        Double optionalThreshold;
        try {
          tempClosestPoi = args.getString(0);
        } catch (JSONException ex) {
          Plog.w("PointrPlugin -> Could not get type of closest poi. Reason: " + ex.getMessage());
          tempClosestPoi = null;
        }
        try {
          optionalTimeout = args.getInt(1);
        } catch (JSONException ex) {
          Plog.w("PointrPlugin -> Could not get timeout. Reason: " + ex.getMessage());
          optionalTimeout = null;
        }

        try {
          optionalThreshold = args.getDouble(2);
        } catch (JSONException ex) {
          Plog.w("PointrPlugin -> Could not get threshold. Reason: " + ex.getMessage());
          optionalThreshold = 5.0;
        }

        typeOfClosestPoi = tempClosestPoi;
        finalOptionalTimeout = optionalTimeout;
        finalOptionalThreshold = optionalThreshold;
        backgroundThread.execute(() -> this.getCurrentLocation(callbackContext, typeOfClosestPoi, finalOptionalTimeout, finalOptionalThreshold));
        return true;
      }
      case "showPathfindingToLocation": {
        final String locationJson;
        try {
          locationJson = args.getString(0);
        } catch (JSONException e) {
          callbackContext.error(e.getMessage());
          return true;
        }
        if (locationJson == null) {
          Plog.w("PointrPlugin ->  Location json is not passed. Cannot take user to location.");
          callbackContext.error("PointrPlugin ->  Location json is not passed");
          return true;
        }
        backgroundThread.execute(() -> this.showPathfindingToLocation(locationJson, callbackContext));
        return true;
      }
      case "setPathfindingMode": {
        final int pathfindingMode;
        try {
          pathfindingMode = args.getInt(0);
        } catch (JSONException e) {
          callbackContext.error(e.getMessage());
          return true;
        }
        backgroundThread.execute(() -> this.setPathfindingMode(pathfindingMode));
        return true;
      }
      case "shouldHideRouteSummaryView": {
        final boolean shouldHide;
        try {
          shouldHide = args.getBoolean(0);
        } catch (JSONException e) {
          callbackContext.error(e.getMessage());
          return true;
        }
        MapActivity.Companion.setRouteSummaryHidden(shouldHide);
        return true;
      }
      case "loadDataForSite": {
        final String siteExternalIdentifier;
        try {
          siteExternalIdentifier = args.getString(0);
        } catch (JSONException e) {
          callbackContext.error(e.getMessage());
          return true;
        }
        backgroundThread.execute(() -> this.loadDataForSite(siteExternalIdentifier, callbackContext));
        return true;
      }
    }
    return false;
  }

  private void loadDataForSite(String siteExternalIdentifier, CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);
  }

  private void setPathfindingMode(Integer pathfindingMode) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        Pointr pointr = Pointr.getPointr();
        PathManager pathManager = pointr.getPathManager();
        PathFindingMode mode;
        if (pathfindingMode == 0) {
          mode = PathFindingMode.NORMAL;
        } else if (pathfindingMode == 1) {
          mode = PathFindingMode.EASY;
        } else if (pathfindingMode == 2) {
          mode = PathFindingMode.ACCESSIBLE;
        } else {
          mode = PathFindingMode.NORMAL;
        }
        pathManager.setCurrentMode(mode);
      }

      @Override
      public void onPointrFailed(String error) {

      }
    }, false, "");
  }

  private void clearPath() {
    cordova.getActivity().runOnUiThread(() -> {
      Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
      mapActivityIntent.setAction(MapActivity.CLEAR_PATH);
      cordova.getActivity().startActivity(mapActivityIntent);
    });
  }

  private void showPathBetweenPois(String siteExternalIdentifier, String sourcePoiInternalIdentifier, String destinationPoiInternalIdentifier, CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        cordova.getActivity().runOnUiThread(() -> {
          Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
          mapActivityIntent.setAction(MapActivity.SHOW_PATHFINDING_BETWEEN_POIS);
          mapActivityIntent.putExtra(MapActivity.site_id_extra_data, siteExternalIdentifier);
          mapActivityIntent.putExtra(MapActivity.source_poi_extra_data, sourcePoiInternalIdentifier);
          mapActivityIntent.putExtra(MapActivity.destination_poi_extra_data, destinationPoiInternalIdentifier);
          cordova.getActivity().startActivity(mapActivityIntent);
        });
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);
  }

  private void initializePointrForClient(String clientId, String licenseKey, String baseUrl, Integer logLevel) {
    PTRParams params = new PTRParams(clientId, licenseKey, baseUrl);
    Plog.LogLevel[] logLevels = Plog.LogLevel.values();
    if (logLevel < 0 || logLevel > logLevels.length - 1) {
      Plog.e("PointrPlugin -> Using default log level INFO");
      logLevel = 2;
    }
    params.setLogLevel(logLevels[logLevel]);
    Pointr.with(cordova.getActivity(), params);
  }

  private void initializePointrForEnv(String env, String licenseKey, Integer logLevel) {
    Plog.LogLevel pointrLogLevel;
    PointrEnvironment environment;
    try {
      environment = PointrEnvironment.valueOf(env.toUpperCase());
      pointrLogLevel = Plog.LogLevel.values()[logLevel];
    } catch (IllegalArgumentException ex) {
      Plog.e("PointrPlugin ->  Critical Error!! Pointr is not initialized");
      return;
    }

    PTRParams params = new PTRParams();
    params.setLicenseKey(licenseKey);
    params.setEnvironment(environment);
    params.setLogLevel(pointrLogLevel);
    Pointr.with(cordova.getActivity(), params);
  }


  /**
   * This method ensures;
   * - Pointr is Running
   * - Data management is complete
   */
  private void ensurePointrReady(PointrReadyCallback pointrReadyCallback, Boolean shouldWaitForDataManagement, String siteExternalIdentifier) {
    Pointr pointr = Pointr.getPointr();
    if (pointr == null) {
      if (clientId != null && !clientId.isBlank() && baseUrl != null && !baseUrl.isBlank()) {
        initializePointrForClient(clientId, licenseKey, baseUrl, logLevel);
      } else {
        initializePointrForEnv(environment, licenseKey, logLevel);
      }
      pointr = Pointr.getPointr();
      if (pointr == null) {
        Plog.e("PointrPlugin ->  Critical Error!! Pointr could not be initialized.");
        pointrReadyCallback.onPointrFailed("PointrPlugin ->  Critical Error!! Pointr could not be initialized.");
        return;
      }
    }
    final Pointr pointrSafe = pointr;

    PointrListener listener = new PointrListener() {
      @Override
      public void onStateUpdated(Pointr.State state) {
        Plog.v("PointrPlugin -> Pointr state updated: " + state.name());
        if (state == Pointr.State.RUNNING) {
          ensurePermissionsAreGiven();
          pointrSafe.removeListener(this);
          if (shouldWaitForDataManagement) {
            ensureDataManagementIsComplete(pointrReadyCallback, siteExternalIdentifier);
          } else {
            pointrReadyCallback.onPointrReady();
          }
        } else if (state.getVal() < 0) {
          pointrSafe.removeListener(this);
          pointrReadyCallback.onPointrFailed("PointrPlugin ->  Critical!! Pointr could not be started. All subsequent calls to Pointr methods will fail. Reason: " + state.name());
        }
      }
    };
    pointrSafe.addListener(listener);
    if (pointrSafe.getState() == Pointr.State.RUNNING) {
      ensurePermissionsAreGiven();
      pointrSafe.removeListener(listener);
      Plog.v("PointrPlugin -> Pointr is already running.");
      if (shouldWaitForDataManagement) {
        ensureDataManagementIsComplete(pointrReadyCallback, siteExternalIdentifier);
      } else {
        pointrReadyCallback.onPointrReady();
      }
    } else {
      pointrSafe.start();
      // Wait for pointrListener
    }
  }

  private void ensurePermissionsAreGiven() {
    if (!shouldRequestPermissionsAtStartup) {
      return;
    }
    permissionManager = Pointr.getPointr().getPermissionManager();
    if (!permissionManager.getHasLocationPermissionWhileInUse()) {
      // request permissions 1 at a time
      cordova.requestPermissions(this, 1, new String[]{Manifest.permission.ACCESS_FINE_LOCATION});
    } else if (!permissionManager.getHasBluetoothPermission()) {
      cordova.requestPermissions(this, 2, new String[]{"android.permission.BLUETOOTH_CONNECT",
        "android.permission.BLUETOOTH_SCAN"});
    }
  }

  @Override
  public void onRequestPermissionResult(int requestCode, String[] permissions, int[] grantResults) throws JSONException {
    super.onRequestPermissionResult(requestCode, permissions, grantResults);
    if (requestCode == 1) {
      if (!permissionManager.getHasBluetoothPermission()) {
        cordova.requestPermissions(this, 2, new String[]{"android.permission.BLUETOOTH_CONNECT",
          "android.permission.BLUETOOTH_SCAN"});
      }
      Plog.v("PointrPlugin ->  Location permission requests has been resulted");
    } else if (requestCode == 2) {
      Plog.v("PointrPlugin ->  Bluetooth permission requests has been resulted");
    }
  }

  private void ensureDataManagementIsComplete(PointrReadyCallback pointrReadyCallback, String siteExternalId) {
    ensureDataManagementIsComplete(pointrReadyCallback, siteExternalId, dataManagementWaitTimeoutInSeconds);
  }

  private void ensureDataManagementIsComplete(PointrReadyCallback pointrReadyCallback, String siteExternalId, Integer timeoutInSeconds) {
    Pointr pointr = Pointr.getPointr();
    DataManager dataManager = pointr.getDataManager();
    SiteManager siteManager = pointr.getSiteManager();
    Site site = siteManager.getSite(siteExternalId);
    Semaphore semaphore = new Semaphore(1);
    SiteManager.Listener siteManagerListener = semaphore::release;
    siteManager.addListener(siteManagerListener);
    final AtomicBoolean didInvokeCallback = new AtomicBoolean(false);
    if (site == null) {
      try {
        semaphore.acquire();
        semaphore.tryAcquire(4L, TimeUnit.SECONDS);
        semaphore.release();
      } catch (InterruptedException e) {

      }
      site = siteManager.getSite(siteExternalId);
    }
    if (site == null) {
      Plog.w("PointrPlugin -> Could not load the sites, so cannot load data");
      pointrReadyCallback.onPointrFailed("Could not load the sites");
      return;
    }
    siteManager.removeListener(siteManagerListener);
    final Site requestedSite = site;
    final Semaphore dataManagementComplete = new Semaphore(0);
    DataManager.Listener dataManagerListener = new DataManager.Listener() {
      @Override
      public void onDataManagerStartDataManagementForSite(@NonNull Site site, boolean b) {
        Plog.v("PointrPlugin -> onDataManagerStartDataManagementForSite: " + site.getTitle());
      }

      @Override
      public void onDataManagerBeginProcessingDataForSite(@NonNull Site site, @Nullable DataType dataType, boolean b) {
        String dataTypeName = "";
        if (dataType != null) {
          dataTypeName = dataType.name();
        }
        Plog.v("PointrPlugin -> onDataManagerBeginProcessingDataForSite - data: " + dataTypeName);
      }

      @Override
      public void onDataManagerEndProcessingDataForSite(@NonNull Site site, @Nullable DataType dataType, boolean b, boolean b1, @Nullable List<ErrorMessage> list) {
        String dataTypeName = "";
        if (dataType != null) {
          dataTypeName = dataType.name();
        }
        Plog.v("PointrPlugin -> onDataManagerEndProcessingDataForSite - data: " + dataTypeName);
      }

      @Override
      public void onDataManagerCompleteAllForSite(@NonNull Site site, boolean b, boolean b1, @Nullable List<ErrorMessage> list) {
        Plog.v("PointrPlugin -> onDataManagerCompleteAllForSite site:" + site.getTitle() + " - isOnline: " + b + " - isSuccess: " + b1);
        if (site.hasSameInternalIdentifier(requestedSite) && b && b1) {
          if (dataManager != null) {
            dataManager.removeListener(this);
          }
          if (didInvokeCallback.compareAndSet(false, true)) {
            pointrReadyCallback.onPointrReady();
          }
          dataManagementComplete.release();
        } else if (site.hasSameInternalIdentifier(requestedSite) && b && !b1) {
          if (dataManager != null) {
            dataManager.removeListener(this);
          }
          String errorList = "";
          if (list != null) {
            errorList = Arrays.toString(list.toArray());
          }
          if (didInvokeCallback.compareAndSet(false, true)) {
            pointrReadyCallback.onPointrFailed(errorList);
          }
          dataManagementComplete.release();
        }
      }

      @Override
      public void onDataManagerReadyForSite(@NonNull Site site) {
        Plog.v("PointrPlugin -> onDataManagerReadyForSite");
        if (site.hasSameInternalIdentifier(requestedSite)) {
          // Site is ready
          if (dataManager != null) {
            dataManager.removeListener(this);
          }
          if (didInvokeCallback.compareAndSet(false, true)) {
            pointrReadyCallback.onPointrReady();
          }
          dataManagementComplete.release();
        }
      }
    };

    dataManager.addListener(dataManagerListener);
    dataManager.loadDataForSite(requestedSite);

    if (dataManager.isSiteContentReady(requestedSite)) {
      // Site is ready
      Plog.i("PointrPlugin -> onSiteContentReady for: " + requestedSite.getTitle());
      dataManager.removeListener(dataManagerListener);
      if (didInvokeCallback.compareAndSet(false, true)) {
        pointrReadyCallback.onPointrReady();
      }
    } else {
      // Wait for dataManager onCompleteAll
      Plog.i("PointrPlugin -> will wait for data management for: " + requestedSite.getTitle());
      boolean didDataManagementComplete = false;
      try {
        didDataManagementComplete = dataManagementComplete.tryAcquire(timeoutInSeconds, TimeUnit.SECONDS);
      } catch (InterruptedException e) {
        Plog.w("Exception while trying to acquire data management complete. ");
      }
      if (!didDataManagementComplete) {
        Plog.w("PointrPlugin -> Could not complete data management. But will continue since this can be due to missing intentional data.");
        if (didInvokeCallback.compareAndSet(false, true)) {
          pointrReadyCallback.onPointrReady();
        }
      }
    }
  }


  private void startPointr(CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, false, "");

  }


  private void stopPointr() {
    Pointr.getPointr().stop();
  }

  // region Action methods

  private void showPoiDetails(String siteExternalIdentifier, String poiInternalIdentifier, CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        openPointrMapActivityWithPoiDetails(siteExternalIdentifier, poiInternalIdentifier);
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);
  }

  private void openPointrMapActivityWithPoiDetails(String siteExternalIdentifier, String poiInternalIdentifier) {
    cordova.getActivity().runOnUiThread(() -> {
      Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
      mapActivityIntent.setAction(MapActivity.SHOW_POI_DETAILS);
      mapActivityIntent.putExtra(MapActivity.site_id_extra_data, siteExternalIdentifier);
      mapActivityIntent.putExtra(MapActivity.destination_poi_extra_data, poiInternalIdentifier);
      cordova.getActivity().startActivity(mapActivityIntent);
    });
  }

  private void showPathfindingToPoi(String siteExternalIdentifier, String poiInternalIdentifier, CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        openPointrMapActivityWithPathfindingToPoi(siteExternalIdentifier, poiInternalIdentifier);
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);
  }

  private void openPointrMapActivityWithPathfindingToPoi(String siteExternalIdentifier, String poiInternalIdentifier) {
    cordova.getActivity().runOnUiThread(() -> {
      Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
      mapActivityIntent.setAction(MapActivity.SHOW_PATHFINDING_TO_POI);
      mapActivityIntent.putExtra(MapActivity.site_id_extra_data, siteExternalIdentifier);
      mapActivityIntent.putExtra(MapActivity.destination_poi_extra_data, poiInternalIdentifier);
      cordova.getActivity().startActivity(mapActivityIntent);
    });
  }

  private void showMapForSite(String siteExternalIdentifier, CallbackContext callbackContext) {
    final String siteExtId = siteExternalIdentifier;
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        cordova.getActivity().runOnUiThread(() -> {
          Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
          mapActivityIntent.setAction(MapActivity.SHOW_SITE);
          mapActivityIntent.putExtra(MapActivity.site_id_extra_data, siteExtId);
          cordova.getActivity().startActivity(mapActivityIntent);
        });
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);
  }

  private void showMapForBuilding(String siteExternalIdentifier, String buildingExternalIdentifier, CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        cordova.getActivity().runOnUiThread(() -> {
          Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
          mapActivityIntent.setAction(MapActivity.SHOW_BUILDING);
          mapActivityIntent.putExtra(MapActivity.site_id_extra_data, siteExternalIdentifier);
          mapActivityIntent.putExtra(MapActivity.building_id_extra_data, buildingExternalIdentifier);
          cordova.getActivity().startActivity(mapActivityIntent);
        });
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);

  }

  private void showMapForLevel(String siteExternalIdentifier, String buildingExternalIdentifier, Integer levelIndex, CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        cordova.getActivity().runOnUiThread(() -> {
          Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
          mapActivityIntent.setAction(MapActivity.SHOW_LEVEL);
          mapActivityIntent.putExtra(MapActivity.site_id_extra_data, siteExternalIdentifier);
          mapActivityIntent.putExtra(MapActivity.building_id_extra_data, buildingExternalIdentifier);
          mapActivityIntent.putExtra(MapActivity.level_id_extra_data, levelIndex);
          cordova.getActivity().startActivity(mapActivityIntent);
        });
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);

  }

  private final int positionWaitTimeoutInSeconds = 60;
  private final int dataManagementWaitTimeoutInSeconds = 60;

  private void getCurrentLocation(CallbackContext callbackContext, String typeOfClosestPoi, Integer timeoutInSeconds, Double thresholdInMeters) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        Pointr pointr = Pointr.getPointr();
        PositionManager positionManager = pointr.getPositionManager();
        CalculatedLocation currentLocation = positionManager.getCurrentCalculatedLocation();
        if (currentLocation == null || !currentLocation.isGeoValid() || currentLocation.getSite() == null) {
          Semaphore positionSemaphore = new Semaphore(0);
          PositionManager.Listener positionListener = new PositionManager.Listener() {
            @Override
            public void onPositionManagerDetectedPositionLevelChange(@NonNull Level level) {
            }

            @Override
            public void onPositionManagerPositionIsFading() {
            }

            @Override
            public void onPositionManagerPositionIsLost() {
            }

            @Override
            public void onPositionManagerPositioningServiceStateChangedTo(PositioningTypes.PositioningServiceState positioningServiceState) {
            }

            @Override
            public void onPositionManagerCalculatedLocation(CalculatedLocation calculatedLocation) {
              if (!calculatedLocation.isGeoValid() || calculatedLocation.getSite() == null) return;
              Pointr.getPointr().getPositionManager().removeListener(this);
              positionSemaphore.release();
            }
          };
          positionManager.addListener(positionListener);
          boolean isAcquired = false;
          int timeout = positionWaitTimeoutInSeconds;
          if (timeoutInSeconds != null) {
            timeout = timeoutInSeconds;
          }
          Plog.i("PointrPlugin -> Position not available yet. Will wait for position manager to heat for: " + timeout + " seconds.");
          try {
            isAcquired = positionSemaphore.tryAcquire(timeout, TimeUnit.SECONDS);
          } catch (InterruptedException ignored) {
          }
          if (!isAcquired) {
            Pointr.getPointr().getPositionManager().removeListener(positionListener);
          }
          currentLocation = positionManager.getCurrentCalculatedLocation();
        }
        if (currentLocation == null || !currentLocation.isGeoValid()) {
          callbackContext.error("PointrPlugin ->  Current user location is not valid. Cannot return location.");
          return;
        }
        Site site = currentLocation.getSite();
        if (site == null) {
          Plog.w("PointrPlugin -> Site is not available. Cannot relate the location to a POI.");
          sendLocationJsonToIonic("", "", currentLocation, callbackContext);
          return;
        }
        final String siteExternalIdentifier = site.getExternalIdentifier();
        if (siteExternalIdentifier.isBlank()) {
          Plog.w("PointrPlugin -> Site external identifier is not available. Cannot relate the location to a POI.");
          sendLocationJsonToIonic("", "", currentLocation, callbackContext);
          return;
        }
        final CalculatedLocation finalLocation = currentLocation;
        ensureDataManagementIsComplete(new PointrReadyCallback() {
          @Override
          public void onPointrReady() {
            Poi poi = getClosestPoi(finalLocation, typeOfClosestPoi, thresholdInMeters);
            if (poi != null) {
              sendLocationJsonToIonic(poi.getName(), poi.getId(), finalLocation, callbackContext);
            } else {
              sendLocationJsonToIonic("", "", finalLocation, callbackContext);
            }
          }

          @Override
          public void onPointrFailed(String error) {
            Plog.w("PointrPlugin -> Could not load data for site: " + siteExternalIdentifier + " due to: " + error + ". Will not return closest poi.");
            sendLocationJsonToIonic("", "", finalLocation, callbackContext);
          }
        }, siteExternalIdentifier, timeoutInSeconds);
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, false, "");

  }

  private Poi getClosestPoi(CalculatedLocation currentLocation, String typeOfClosestPoi, double threshold) {
    Pointr pointr = Pointr.getPointr();
    if (pointr == null) {
      Plog.w("PointrPlugin -> Pointr is not available!");
      return null;
    }
    PoiManager poiManager = pointr.getPoiManager();
    if (poiManager == null) {
      Plog.w("PointrPlugin -> PoiManager not available");
      return null;
    }
    if (currentLocation == null) {
      Plog.w("PointrPlugin -> Location is null");
      return null;
    }
    if (currentLocation.getSite() == null) {
      Plog.w("PointrPlugin -> No Site in location");
      return null;
    }
    GeoPoint point = currentLocation.getGeoPoint();
    if (point == null) {
      Plog.w("PointrPlugin -> No GeoPoint in location");
      return null;
    }
    List<String> listOfTypes = new ArrayList<>();
    if (typeOfClosestPoi != null && !typeOfClosestPoi.isBlank()) {
      listOfTypes.add(typeOfClosestPoi);
    }
    PoiContainer poiContainer;
    if (currentLocation.getLevel() != null) {
      poiContainer = poiManager.getPois(currentLocation.getLevel(), listOfTypes, true);
    } else {
      poiContainer = poiManager.getPois(currentLocation.getSite(), listOfTypes, true);
    }
    if (poiContainer == null) {
      Plog.w("PointrPlugin -> Could not get PoiContainer");
      return null;
    }
    List<Poi> pois = poiContainer.getPoiList();
    if (pois.isEmpty()) {
      Plog.w("PointrPlugin -> No Pois inside PoiContainer");
      return null;
    }
    for (Poi poi : pois) {
      if (poi.getLocation().getLevel() == null) continue;
      Geometry geometry = poi.getGeometry();
      if (geometry == null) continue;
      GeometryType type = geometry.getType();
      if (type == null) continue;
      if (type == GeometryType.Point) continue;
      if (!geometry.isInside(point)) continue;
      return poi;
    }
    PathManager pathManager = pointr.getPathManager();
    if (pathManager == null) {
      Plog.w("PointrPlugin -> PathManager is null, will return first poi");
      return pois.get(0);
    }
    double distance = pathManager.getApproximateDistance(currentLocation, pois.get(0));
    if (distance > (threshold + 1)) {
      Plog.w("PointrPlugin -> distance(" + distance + ") is greater than threshold(" + threshold + "), will return null");
      return null;
    }
    return pois.get(0);
  }


  private void sendLocationJsonToIonic(String poiTitle, String poiInternalIdentifier, CalculatedLocation currentLocation, CallbackContext callbackContext) {
    Building building = currentLocation.getBuilding();
    String buildingExternalIdentifier = "";
    if (building != null) {
      buildingExternalIdentifier = building.getExternalIdentifier();
    }
    Site site = currentLocation.getSite();
    String siteExternalIdentifier = "";
    if (site != null) {
      siteExternalIdentifier = site.getExternalIdentifier();
    }
    Level level = currentLocation.getLevel();
    int levelId = -9999;
    if (level != null) {
      levelId = level.getIndex();
    }
    Location location = new Location(currentLocation.getLat(),
      currentLocation.getLon(),
      levelId,
      buildingExternalIdentifier,
      siteExternalIdentifier,
      poiInternalIdentifier,
      poiTitle
    );
    String locationString = new Gson().toJson(location);
    Plog.v("PointrPlugin -> Sending location back: " + locationString);
    callbackContext.success(locationString);
  }

  private void showPathfindingToLocation(String locationJson, CallbackContext callbackContext) {
    Location location = new Gson().fromJson(locationJson, Location.class);
    if (location == null) {
      callbackContext.error("PointrPlugin ->  Could not get Location");
      return;
    }
    if (location.getSiteExternalIdentifier().isEmpty()) {
      callbackContext.error("PointrPlugin -> Could not get site for location");
      return;
    }
    if (location.getBuildingExternalIdentifier().isEmpty()) {
      callbackContext.error("PointrPlugin -> Could not get building for location");
      return;
    }

    if (location.getPoiInternalIdentifier().isEmpty()) {
      // Means this is an arbitrary location to pathfind to
      ensurePointrReady(new PointrReadyCallback() {
        @Override
        public void onPointrReady() {
          cordova.getActivity().runOnUiThread(() -> {
            Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
            mapActivityIntent.setAction(MapActivity.SHOW_PATHFINDING_TO_LOCATION);
            mapActivityIntent.putExtra(MapActivity.location_extra_data, locationJson);
            cordova.getActivity().startActivity(mapActivityIntent);
          });
          callbackContext.success();
        }

        @Override
        public void onPointrFailed(String error) {
          callbackContext.error(error);
        }
      }, false, "");
    } else {
      // Means we should pathfind to a poi or within a site
      ensurePointrReady(new PointrReadyCallback() {
        @Override
        public void onPointrReady() {
          cordova.getActivity().runOnUiThread(() -> {
            Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
            if (!location.getPoiInternalIdentifier().isEmpty()) {
              mapActivityIntent.setAction(MapActivity.SHOW_PATHFINDING_TO_POI);
              mapActivityIntent.putExtra(MapActivity.destination_poi_extra_data, location.getPoiInternalIdentifier());
              mapActivityIntent.putExtra(MapActivity.site_id_extra_data, location.getSiteExternalIdentifier());
            } else {
              mapActivityIntent.setAction(MapActivity.SHOW_PATHFINDING_TO_LOCATION);
              mapActivityIntent.putExtra(MapActivity.location_extra_data, locationJson);
            }
            cordova.getActivity().startActivity(mapActivityIntent);
          });
          callbackContext.success();
        }

        @Override
        public void onPointrFailed(String error) {
          callbackContext.error(error);
        }
      }, true, location.getSiteExternalIdentifier());
    }


  }

  private void highlightPoi(String siteExternalIdentifier, String poiExternalIdentifier, CallbackContext callbackContext) {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        cordova.getActivity().runOnUiThread(() -> {
          Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
          mapActivityIntent.setAction(MapActivity.HIGHLIGHT_POI);
          mapActivityIntent.putExtra(MapActivity.site_id_extra_data, siteExternalIdentifier);
          mapActivityIntent.putExtra(MapActivity.poi_id_extra_data, poiExternalIdentifier);
          cordova.getActivity().startActivity(mapActivityIntent);
        });
        callbackContext.success();
      }

      @Override
      public void onPointrFailed(String error) {
        callbackContext.error(error);
      }
    }, true, siteExternalIdentifier);
  }

  private void showMap() {
    ensurePointrReady(new PointrReadyCallback() {
      @Override
      public void onPointrReady() {
        Intent mapActivityIntent = new Intent(cordova.getContext(), MapActivity.class);
        mapActivityIntent.setAction(MapActivity.SHOW_MAP);
        cordova.getActivity().runOnUiThread(() -> cordova.getActivity().startActivity(mapActivityIntent));
      }

      @Override
      public void onPointrFailed(String error) {

      }
    }, false, "");
  }

  interface PointrReadyCallback {
    void onPointrReady();

    void onPointrFailed(String error);
  }
}
