package com.pointrlabs.cordova.pointr.core.activity

import android.content.Context
import android.content.Intent
import android.os.Bundle
import android.os.Handler
import android.widget.TextView
import android.widget.Toast
import androidx.appcompat.app.AppCompatActivity
import capacitor.cordova.android.plugins.R
import com.google.gson.Gson
import com.pointrlabs.core.geometry.GeoPoint
import com.pointrlabs.core.management.*
import com.pointrlabs.core.management.interfaces.PointrListener
import com.pointrlabs.core.management.models.Building
import com.pointrlabs.core.management.models.Site
import com.pointrlabs.core.map.handlers.TrackingModeEventsHandler
import com.pointrlabs.core.map.models.events_listeners.MapEventsListener
import com.pointrlabs.core.map.views.PTRMapFragment
import com.pointrlabs.core.map.views.PTRMapWidgetFragment
import com.pointrlabs.core.nativecore.wrappers.Plog
import com.pointrlabs.core.pathfinding.PathManager
import com.pointrlabs.core.positioning.model.CalculatedLocation
import com.pointrlabs.core.site.SiteManager
import com.pointrlabs.cordova.pointr.core.models.Location
import com.pointrlabs.core.map.handlers.ExitButtonEventsHandler
import com.pointrlabs.core.map.models.PTRMapViewMarker
import com.pointrlabs.core.map.viewmodels.PTRMapWidgetConfiguration
import java.util.concurrent.Executors
import android.content.DialogInterface
import android.app.AlertDialog
import androidx.lifecycle.Lifecycle


/**
 * Accepts the facility and displays the map with given facility
 */
class MapActivity : AppCompatActivity(), PointrListener {

  companion object {
    private const val BUNDLE_FACILITY_ID = "BUNDLE_FACILITY_ID"
    const val HIGHLIGHT_POI = "ACTION_HIGHLIGHT_POI"
    const val SHOW_POI_DETAILS = "ACTION_SHOW_POI_DETAILS"
    const val SHOW_PATHFINDING_TO_POI = "ACTION_SHOW_ROUTE_TO_POI"
    const val SHOW_PATHFINDING_BETWEEN_POIS = "ACTION_SHOW_ROUTE_FROM_TO"
    const val SHOW_SITE = "ACTION_SHOW_SITE"
    const val SHOW_BUILDING = "ACTION_SHOW_BUILDING"
    const val SHOW_LEVEL = "ACTION_SHOW_LEVEL"
    const val CLEAR_PATH = "ACTION_CLEAR_PATH"
    const val SHOW_PATHFINDING_TO_LOCATION = "ACTION_SHOW_ROUTE_TO_CAR"
    const val SHOW_MAP = "ACTION_SHOW_MAP"
    const val site_id_extra_data = "site_id_extra_data"
    const val building_id_extra_data = "building_id_extra_data"
    const val level_id_extra_data = "level_id_extra_data"
    const val source_poi_extra_data = "source_poi_extra_data"
    const val destination_poi_extra_data = "destination_poi_extra_data"
    const val location_extra_data = "location_extra_data"
    const val poi_id_extra_data = "poi_id_extra_data"
    var shouldHideRouteSummaryView = false
    fun setRouteSummaryHidden(shouldHide: Boolean) {
      shouldHideRouteSummaryView = shouldHide
    }
    fun newIntent(context: Context, facilityId: Int) =
      Intent(context, MapActivity::class.java).apply {
        putExtra(BUNDLE_FACILITY_ID, facilityId)
      }
  }


  val backgroundThread = Executors.newFixedThreadPool(1)

  private var mapWidgetFragment: PTRMapWidgetFragment? = null
  private var pendingAction: () -> Unit = {}
  private lateinit var pointr: Pointr
  private lateinit var poiManager: PoiManager
  private lateinit var siteManager: SiteManager
  private lateinit var positionManager: PositionManager
  private lateinit var pathManager: PathManager
  var handler: Handler? = null


  override fun onCreate(savedInstanceState: Bundle?) {
    super.onCreate(savedInstanceState)
    setContentView(R.layout.activity_map)
    pointr = Pointr.getPointr()!!
    if (pointr.state != Pointr.State.RUNNING) {
      Plog.w(
        "PointrPlugin ->  Pointr must be running for this activity to be started. Please call Pointr.start" +
          "methods before using this."
      )
      return
    }
    poiManager = pointr.poiManager!!
    siteManager = pointr.siteManager!!
    positionManager = pointr.positionManager!!
    pathManager = pointr.pathManager!!

    // Create map widget fragment
    val configuration = PTRMapWidgetConfiguration.defaultConfiguration()
    configuration.isRouteSummaryEnabled = !shouldHideRouteSummaryView
    mapWidgetFragment =
      PTRMapWidgetFragment.newInstance(configuration)
        .show(supportFragmentManager, R.id.fragment_container)

    mapWidgetFragment?.exitButtonEventsHandler = object : ExitButtonEventsHandler() {
          override fun onExitClicked(mapWidget: PTRMapWidgetFragment, exitButton: TextView) {
              val activity = mapWidget.activity ?: return
              if (!activity.lifecycle.currentState.isAtLeast(Lifecycle.State.RESUMED)) return
              AlertDialog.Builder(activity)
                  .setMessage("Do you want to exit?")
                  .setPositiveButton("Yes") { dialog, _ ->
                      activity.finish()
                      dialog?.dismiss()
                  }
                  .setNegativeButton("No") { dialog: DialogInterface?, _: Int -> dialog?.dismiss() }
                  .show()

          }
      }
    mapWidgetFragment?.addListener(object : MapEventsListener {
      override fun mapDidEndLoading(mapFragment: PTRMapFragment) {
        mapFragment.maximumZoomLevel = 20.0
      }
    })

    handler = Handler(mainLooper)
    pendingAction = {}
    when (intent.action) {
      HIGHLIGHT_POI -> {
        highlightPoi()
        return
      }
      SHOW_POI_DETAILS -> {
        showPoiDetails()
        return
      }
      SHOW_PATHFINDING_TO_LOCATION -> {
        showPathfindingToLocation()
        return
      }
      SHOW_PATHFINDING_TO_POI -> {
        showPathfindingToPoi()
        return
      }
      SHOW_BUILDING -> {
        showBuildingMap()
        return
      }
      SHOW_SITE -> {
        showSiteMap()
        return
      }
      SHOW_LEVEL -> {
        showLevelMap()
        return
      }
      SHOW_PATHFINDING_BETWEEN_POIS -> {
        showPathfindingBetweenPois()
        return
      }
      CLEAR_PATH -> {
        pendingAction = { clearPathFromMap() }
      }
      else -> {
        Plog.i("PointrPlugin ->  No action set. Will show default map")
        pendingAction = {}
      }
    }
    showMap()
  }

  private fun clearPathFromMap() {
    Plog.v("PointrPlugin ->  clearPathFromMap")
    handler?.post {
      Plog.v("PointrPlugin ->  clearPathFromMap post handler")
      mapWidgetFragment?.mapFragment?.currentPath = null
    }
  }

  private fun showPathfindingBetweenPois() {
    Plog.v("PointrPlugin ->  showPathfindingBetweenPois")
    val siteExternalIdentifier = intent.getStringExtra(site_id_extra_data)
    if (siteExternalIdentifier == null) {
      Plog.w("PointrPlugin ->  Could not get site external identifier.")
      finish()
      return
    }
    val site = siteManager.getSite(siteExternalIdentifier)
    if (site == null) {
      Plog.w("PointrPlugin ->  Cannot find site with external id: $siteExternalIdentifier")
      mapWidgetFragment?.showSite(siteExternalIdentifier)
      return
    }
    val sourcePoiExtra = intent.getStringExtra(source_poi_extra_data)
    if (sourcePoiExtra == null) {
      Plog.w("PointrPlugin ->  Could not get source poi extra data. Cannot start static route")
      mapWidgetFragment?.showSite(site)
      return
    }
    val sourcePoi = poiManager.getPoiByIdentifier(site, sourcePoiExtra)
    if (sourcePoi == null) {
      Plog.w("PointrPlugin ->  Could not find Poi with identifier: $sourcePoiExtra - Cannot start static route")
      mapWidgetFragment?.showSite(site)
      return
    }
    val destinationPoiExtra = intent.getStringExtra(destination_poi_extra_data)
    if (destinationPoiExtra == null) {
      Plog.w("PointrPlugin ->  Could not get destination poi extra data. Cannot start static route")
      mapWidgetFragment?.showSite(site)
      return
    }
    val destinationPoi = poiManager.getPoiByIdentifier(site, destinationPoiExtra)
    if (destinationPoi == null) {
      Plog.w("PointrPlugin ->  Could not find Poi with identifier: $destinationPoiExtra - cannot start static route")
      mapWidgetFragment?.showSite(site)
      return
    }
    val path = pathManager.calculatePath(sourcePoi, listOf(destinationPoi))
    if (path == null) {
      Plog.w("PointrPlugin ->  Could not find path between source: ${sourcePoi.name} and destination: ${destinationPoi.name}")
      mapWidgetFragment?.showSite(site)
      return
    }
    mapWidgetFragment?.exitButtonEventsHandler = object : ExitButtonEventsHandler() {
      override fun onExitClicked(mapWidget: PTRMapWidgetFragment, exitButton: TextView) {
        Plog.v("PointrPlugin ->  showPathfindingBetweenPois exit button clicked")
        val activity = mapWidget.activity ?: return
        if (!activity.lifecycle.currentState.isAtLeast(Lifecycle.State.RESUMED)) return
        AlertDialog.Builder(activity)
          .setMessage("Do you want to exit?")
          .setPositiveButton("Yes") { dialog, _ ->
            handler?.post {
              mapWidgetFragment?.removeAllAnnotations()
              mapWidgetFragment?.mapFragment?.currentPath = null
            }
            activity.finish()
            dialog?.dismiss()
          }
          .setNegativeButton("No") { dialog: DialogInterface?, _: Int -> dialog?.dismiss() }
          .show()
      }
    }
    if (handler == null) {
      handler = Handler(mainLooper)
    }
    handler?.post {
      Plog.v("PointrPlugin ->  showPathfindingBetweenPois post handler")
      mapWidgetFragment?.showLevel(sourcePoi.location.level!!) {
        handler?.postDelayed({
          Plog.v("PointrPlugin ->  showPathfindingBetweenPois post delayed handler")
          mapWidgetFragment?.addAnnotation(
            PTRMapViewMarker(
              latitude = sourcePoi.location.lat,
              longitude = sourcePoi.location.lon,
              text = sourcePoi.name,
              shouldAllowOverlap = true
            ), sourcePoi.location.level!!
          )
          Plog.v("PointrPlugin ->  showPathfindingBetweenPois added source poi annotation")
          mapWidgetFragment?.addAnnotation(
            PTRMapViewMarker(
              latitude = destinationPoi.location.lat,
              longitude = destinationPoi.location.lon,
              text = destinationPoi.name,
              shouldAllowOverlap = true
            ), destinationPoi.location.level!!
          )
          Plog.v("PointrPlugin ->  showPathfindingBetweenPois added destination poi annotation")
          mapWidgetFragment?.mapFragment?.zoomToLocation(
            GeoPoint(
              sourcePoi.location.lat,
              sourcePoi.location.lon
            )
          )
          Plog.v("PointrPlugin ->  showPathfindingBetweenPois setting current path")
          mapWidgetFragment?.mapFragment?.currentPath = path
        }, 1000L)
      }
    }
  }

  private fun showLevelMap() {
    val siteExternalIdentifier = intent.getStringExtra(site_id_extra_data)
    if (siteExternalIdentifier == null) {
      Plog.w("PointrPlugin ->  Could not get site external identifier.")
      return
    }
    val buildingExternalIdentifier = intent.getStringExtra(building_id_extra_data)
    if (buildingExternalIdentifier == null) {
      Plog.w("PointrPlugin ->  Could not get building external identifier. Will show default level")
      mapWidgetFragment?.showSite(siteExternalIdentifier)
      return
    }
    val levelId = intent.getIntExtra(level_id_extra_data, 0)
    mapWidgetFragment?.showLevel(siteExternalIdentifier, buildingExternalIdentifier, levelId)
  }

  private fun showSiteMap() {
    val siteExternalIdentifier = intent.getStringExtra(site_id_extra_data)
    if (siteExternalIdentifier == null) {
      Plog.w("PointrPlugin ->  Could not get site external identifier.")
      return
    }
    mapWidgetFragment?.showSite(siteExternalIdentifier)
  }

  private fun showBuildingMap() {
    val siteExternalIdentifier = intent.getStringExtra(site_id_extra_data)
    if (siteExternalIdentifier == null) {
      Plog.w("PointrPlugin ->  Could not get site external identifier.")
      finish()
      return
    }
    val buildingExternalIdentifier = intent.getStringExtra(building_id_extra_data)
    if (buildingExternalIdentifier == null) {
      Plog.w("PointrPlugin ->  Could not get building external identifier. Will show default level")
      mapWidgetFragment?.showSite(siteExternalIdentifier)
      return
    }
    mapWidgetFragment?.showBuilding(siteExternalIdentifier, buildingExternalIdentifier)
  }

  private fun showPathfindingToPoi() {
    Plog.v("PointrPlugin ->  showPathfindingToPoi")
    val siteExternalIdentifier = intent.getStringExtra(site_id_extra_data)
    if (siteExternalIdentifier == null) {
      Plog.w("PointrPlugin ->  Could not get site external identifier.")
      finish()
      return
    }
    val site = siteManager.getSite(siteExternalIdentifier)
    if (site == null) {
      Plog.w("PointrPlugin ->  Cannot find site with external id: $siteExternalIdentifier")
      mapWidgetFragment?.showSite(siteExternalIdentifier)
      return
    }
    val destinationPoiExtra = intent.getStringExtra(destination_poi_extra_data)
    if (destinationPoiExtra == null) {
      Plog.w("PointrPlugin ->  Could not get destination poi extra data. Cannot start static route")
      mapWidgetFragment?.showSite(site)
      return
    }
    val destinationPoi = poiManager.getPoiByIdentifier(site, destinationPoiExtra)
    if (destinationPoi == null) {
      Plog.w("PointrPlugin ->  Could not find Poi with identifier: $destinationPoiExtra - cannot start static route")
      mapWidgetFragment?.showSite(site)
      return
    }
    Plog.v("PointrPlugin ->  Calculating path to poi: ${destinationPoi.name}")
    val pathSession = pathManager.startPathSession(listOf(destinationPoi))
    if (pathSession != null) {
      Plog.v("PointrPlugin ->  Successfully calculated path")
      if (handler == null) {
        handler = Handler(mainLooper)
      }
      handler?.post {
        Plog.v("PointrPlugin ->  showPathfindingToPoi post handler")
        mapWidgetFragment?.showPathFinding(pathSession, onComplete = null)
      }
    } else {
      Plog.v("PointrPlugin ->  Could not calculate path to poi: ${destinationPoi.name}")
      handler?.post {
        mapWidgetFragment?.showPoiDetails(destinationPoi)
      }
    }

  }

  private fun showPoiDetails() {
    Plog.v("PointrPlugin ->  showPoiDetails")
    val siteExternalIdentifier = intent.getStringExtra(site_id_extra_data)
    if (siteExternalIdentifier == null) {
      Plog.w("PointrPlugin ->  Could not get site external identifier.")
      finish()
      return
    }
    val site = siteManager.getSite(siteExternalIdentifier)
    if (site == null) {
      Plog.w("PointrPlugin ->  Cannot find site with external id: $siteExternalIdentifier")
      mapWidgetFragment?.showSite(siteExternalIdentifier)
      return
    }
    val destinationPoiExtra = intent.getStringExtra(destination_poi_extra_data)
    if (destinationPoiExtra == null) {
      Plog.w("PointrPlugin ->  Could not get destination poi extra data. Cannot start static route")
      mapWidgetFragment?.showSite(site)
      return
    }
    val destinationPoi = poiManager.getPoiByIdentifier(site, destinationPoiExtra)
    if (destinationPoi == null) {
      Plog.w("PointrPlugin ->  Could not find Poi with identifier: $destinationPoiExtra - cannot start static route")
      mapWidgetFragment?.showSite(site)
      return
    }
    if (handler == null) {
      handler = Handler(mainLooper)
    }
    handler?.post {
      Plog.v("PointrPlugin ->  showPoiDetails post handler")
      mapWidgetFragment?.showPoiDetails(destinationPoi)
    }
  }

  private fun showPathfindingToLocation() {
    Plog.v("PointrPlugin ->  showPathfindingToLocation")
    val carLocationString =
      intent.getStringExtra(location_extra_data) // Get the location of the car passed from ionic side by intent
    if (carLocationString == null) {
      Plog.w(
        "PointrPlugin ->  Location of the car was not given. Cannot start navigation to car. " +
          "Please pass the location of the car"
      )
      finish()
      return
    }
    val carLocation: Location
    try {
      carLocation = Gson().fromJson(carLocationString, Location::class.java)
    } catch (ex: Exception) {
      Plog.w(
        "PointrPlugin ->  Failed to convert given car location json to CarLocation class. Please " +
          "follow the documentation to pass in the correct format."
      )
      Plog.w(ex.message)
      finish()
      return
    }

    val site = siteManager.getSite(carLocation.siteExternalIdentifier)
    if (site == null) {
      Plog.w("PointrPlugin ->  Could not find site with id: ${carLocation.siteExternalIdentifier}")
      mapWidgetFragment?.showSite(carLocation.siteExternalIdentifier)
      return
    }

    val currentLocation = positionManager.currentCalculatedLocation
    if (currentLocation == null || currentLocation.isValid().not()) {
      Plog.w("PointrPlugin ->  The user needs to be onsite to be able to navigate to their car.")
      mapWidgetFragment?.showSite(site)
      return
    }

    val building = siteManager.getBuilding(carLocation.buildingExternalIdentifier)
    if (building == null) {
      Plog.w("PointrPlugin ->  Could not find building with id: ${carLocation.buildingExternalIdentifier}")
      mapWidgetFragment?.showSite(site)
      return
    }

    val level = building.levels.firstOrNull { it.index == carLocation.levelIndex }
    if (level == null) {
      Plog.w("PointrPlugin ->  Could not find level index: ${carLocation.levelIndex} in building: ${building.title}.")
      mapWidgetFragment?.showBuilding(building)
      return
    }

    val pathSession = pathManager.startPathSession(
      listOf(
        CalculatedLocation(
          carLocation.latitude,
          carLocation.longitude,
          site,
          building,
          level
        )
      )
    )
    if (pathSession == null) {
      Plog.w("PointrPlugin ->  Failed to calculate path.")
      mapWidgetFragment?.showLevel(level)
      return
    }
    if (handler == null) {
      handler = Handler(mainLooper)
    }
    handler?.post {
      Plog.v("PointrPlugin ->  showPathfindingToLocation post handler")
      mapWidgetFragment?.showPathFinding(pathSession, onComplete = null)
    }
  }

  private fun highlightPoi() {
    Plog.v("PointrPlugin ->  highlightPoi")
    val siteExternalIdentifier =
      intent.getStringExtra(site_id_extra_data) // Gets the Massport identifier
    if (siteExternalIdentifier == null) {
      Plog.w("PointrPlugin ->  No site identifier passed. Cannot zoom in to poi")
      finish()
      return
    }
    val site = siteManager.getSite(siteExternalIdentifier)
    if (site == null) {
      Plog.w("PointrPlugin ->  No site found with given identifier: $siteExternalIdentifier")
      mapWidgetFragment?.showSite(siteExternalIdentifier)
      return
    }

    val poiId =
      intent.getStringExtra(poi_id_extra_data) // Fetches the Massport Poi id passed from intent
    if (poiId == null) {
      Plog.i("PointrPlugin ->  No Poi Id passed to intent to zoom in to. Please pass a poi id using the $poi_id_extra_data")
      mapWidgetFragment?.showSite(site)
      return
    }
    val matchingPoi = poiManager.getPoiByIdentifier(site, poiId)
    if (matchingPoi != null) {
      if (handler == null) {
        handler = Handler(mainLooper)
      }
      handler?.post {
        Plog.v("PointrPlugin ->  highlightPoi post handler")
        mapWidgetFragment?.showLevel(matchingPoi.location.level!!) {
          mapWidgetFragment?.mapFragment!!.highlightPoi(matchingPoi)
          handler?.postDelayed({
            Plog.v("PointrPlugin ->  highlightPoi delayed post handler")
            mapWidgetFragment?.mapFragment?.zoomToLocation(
              GeoPoint(
                matchingPoi.location.lat,
                matchingPoi.location.lon
              )
            )
          }, 1000L)
        }
      }
    } else {
      Plog.i("PointrPlugin ->  No poi was found with id: $poiId in site: ${site.title}")
      mapWidgetFragment?.showSite(site)
    }
  }


  override fun onDestroy() {
    super.onDestroy()
    Pointr.getPointr()!!.removeListener(this)
    mapWidgetFragment?.onDestroy()
    mapWidgetFragment = null
    Plog.v("PointrPlugin ->  onDestroy")
  }


  override fun onStateUpdated(state: Pointr.State?) {
    handler?.post {
      Toast.makeText(this, state.toString(), Toast.LENGTH_SHORT).show()
    }
    Plog.v("PointrPlugin ->  onStateUpdated: ${state?.name}")
    if (state == Pointr.State.RUNNING) {
      showMap()
    }
  }


  private fun showMap() {
    Plog.v("PointrPlugin ->  showMap")
    handler?.post {
      Plog.v("PointrPlugin ->  showMap post handler")
      val sitesList = siteManager.getSites()
      if (sitesList.isEmpty()) {
        Plog.e("PointrPlugin -> No sites found. Cannot show map")
        finish()
        return@post
      }
      val buildingList = sitesList.first().buildings
      if (buildingList.isEmpty()) {
        Plog.e("PointrPlugin -> No buildings found. Cannot show map")
        mapWidgetFragment?.showSite(sitesList.first())
        return@post
      }
      val building = buildingList.first()
      supportActionBar?.title = building.site?.title
      val mapWidgetFragment = mapWidgetFragment ?: return@post

      mapWidgetFragment.showBuilding(building) {
        Plog.v("PointrPlugin ->  showMap - mapDidEndLoading")
        backgroundThread.submit {
          Plog.v("PointrPlugin ->  invoking pending Action")
          pendingAction.invoke()
          pendingAction = {}
        }
      }

    }
  }
}
